# course_model.py

class Course:
    """Defines a course with relationships and metadata."""

    def __init__(self, name, credits, term=None, completed=False, critical_path=True, categories=None, full_name=None, note=None, ms_credits=None, writing_intensive=False):
        self.name = name
        self.credits = credits
        self.term = term
        self.completed = completed
        self.critical_path = critical_path
        self.prereqs = []
        self.coreqs = []
        self.coprereqs = []
        self.styles = {}
        self.categories = categories
        self.full_name = full_name or name
        self.note = note
        if self.note is None:
            if self.full_name:
                self.note = f"{self.full_name}"
            else:
                self.note = f"{self.name} ({self.credits} cr)"
        if "MS" in self.categories:
            self.ms_credits = ms_credits if ms_credits is not None else credits
        else:
            self.ms_credits = 0
        self.writing_intensive = writing_intensive

    def add_prereq(self, prereq_name):
        self.prereqs.append(prereq_name)
        return self

    def add_coreq(self, coreq_name):
        self.coreqs.append(coreq_name)
        return self

    def add_coprereq(self, coprereq_name):
        self.coprereqs.append(coprereq_name)
        return self

    def add_style(self, style_key, style_value):
        self.styles[style_key] = style_value
        return self

    def set_completed(self, completed=True):
        self.completed = completed
        return self

def new_course(name, credits, term=None, completed=False, categories=None):
    """Creates and registers a new course."""
    course = Course(name, credits, term, completed, categories)
    return course


#!/usr/local/bin/python

import sys
import graphviz as gv
import functools
from pprint import pprint
import random
import math

#### functions and classes ####

def apply_styles(graph, styles):
  graph.graph_attr.update(
    ('graph' in styles and styles['graph']) or {}
  )
  graph.node_attr.update(
    ('nodes' in styles and styles['nodes']) or {}
  )
  graph.edge_attr.update(
    ('edges' in styles and styles['edges']) or {}
  )
  return graph

def add_nodes(graph, nodes):
  global courses
  for n in nodes:
    if isinstance(n, tuple):
      graph.node(courses[n[0]].name+' '+str(courses[n[0]].credits), **n[1])
    else:
      if courses[n].completed:
        graph.attr('node', fillcolor='#cecdc9', color='#cecdc9')
      else:
        graph.attr('node', fillcolor='white', color='white')
      graph.node(courses[n].name,'<<font color="#6e6d6a"><sub>'+str(courses[n].credits)+'</sub></font>   '+courses[n].name+'>')
  return graph

def add_edges(graph, edges):
  for e in edges:
    if isinstance(e[0], tuple):
      graph.edge(*e[0], **e[1])
    else:
      graph.edge(*e)
  return graph

class course:
  """Defines a course"""

  def __init__(self, name, credits, term, completed=False):
    self.name = name
    self.credits = credits
    self.term = term
    self.completed = completed
    self.prereqs = []
    self.coreqs = []
    self.coprereqs = []
    self.styles = {}

  def add_prereq(self, prereq):
    self.prereqs.append(prereq)
    return self

  def add_coreq(self, coreq):
    self.coreqs.append(coreq)
    return self

  def add_coprereq(self, coprereq):
    self.coprereqs.append(coprereq)
    return self

  def add_style(self, style_key, style_value):
    self.styles.update({style_key: style_value})
    return self

  def set_completed(self, coprereq=True):
    self.completed = coprereq
    return self

def new_course(name,credits,term,completed=False):
  global courses
  try: courses
  except NameError: courses = {}
  courses.update({name: course(name,credits,term,completed)})
  return courses[name]

#### colors ####

kelly_colors_hex = [
  0xFFB300, # Vivid Yellow
  0x803E75, # Strong Purple
  0xFF6800, # Vivid Orange
  0xA6BDD7, # Very Light Blue
  0xC10020, # Vivid Red
  0xCEA262, # Grayish Yellow
  0x817066, # Medium Gray

  # The following don't work well for people with defective color vision
  0x007D34, # Vivid Green
  0xF6768E, # Strong Purplish Pink
  0x00538A, # Strong Blue
  0xFF7A5C, # Strong Yellowish Pink
  0x53377A, # Strong Violet
  0xFF8E00, # Vivid Orange Yellow
  0xB32851, # Strong Purplish Red
  0xF4C800, # Vivid Greenish Yellow
  0x7F180D, # Strong Reddish Brown
  0x93AA00, # Vivid Yellowish Green
  0x593315, # Deep Yellowish Brown
  0xF13A13, # Vivid Reddish Orange
  0x232C16, # Dark Olive Green
  ]

#### build the graph ####

graph = functools.partial(gv.Graph, format='svg')
digraph = functools.partial(gv.Digraph, format='svg')

exec(open(sys.argv[1]).read())

nodes_by_term = [[],[],[],[],[],[],[],[]] # initialize
terms_by_name = ['1F','1S','2F','2S','3F','3S','4F','4S']
for course_key in courses.keys():
  course_term_index = terms_by_name.index(courses[course_key].term)
  nodes_by_term[course_term_index].append(courses[course_key].name)

edges = [] # initialize
edges_cluster = [] # initialize
color_lookup = {} # initialize
i_mod = 0
for course_key in courses.keys():
  for prereq in courses[course_key].prereqs:
    if prereq in color_lookup:
      edge_color = color_lookup[prereq]
    else:
      i_mod = (i_mod+1)%len(kelly_colors_hex)
      edge_color = str("#%06x" % kelly_colors_hex[i_mod])
      color_lookup[prereq] = edge_color
    if not courses[prereq].completed:
      edges.append(
        ((prereq,courses[course_key].name),{'style': 'solid','color': edge_color})
      )
  for coreq in courses[course_key].coreqs:
    if coreq in color_lookup:
      edge_color = color_lookup[coreq]
    else:
      i_mod = (i_mod+1)%len(kelly_colors_hex)
      edge_color = str("#%06x" % kelly_colors_hex[i_mod])
      color_lookup[coreq] = edge_color
    if not courses[coreq].completed:
      edges.append(
        ((coreq,courses[course_key].name),{'style': 'solid','color': edge_color,'dir': 'both'})
      )
  for coprereq in courses[course_key].coprereqs:
    if coprereq in color_lookup:
      edge_color = color_lookup[coprereq]
    else:
      i_mod = (i_mod+1)%len(kelly_colors_hex)
      edge_color = str("#%06x" % kelly_colors_hex[i_mod])
      color_lookup[coprereq] = edge_color
    if not courses[coprereq].completed:
      edges.append(
        ((coprereq,courses[course_key].name),{'style': 'dashed','color': edge_color})
      )

g = digraph(engine='dot')
g.attr(newrank='true',compound='true')
for i,term in enumerate(terms_by_name):
  term_credits = 0
  for c in nodes_by_term[i]:
    term_credits += courses[c].credits
  name_now = 'cluster_'+str(i)
  label_cluster = {'label': '<<B>term: '+str(terms_by_name[i])+'</B><BR/><FONT POINT-SIZE="10" COLOR="#6e6d6a">'+str(term_credits)+' cr</FONT>>'}
  h = add_nodes(digraph(name_now), nodes_by_term[i])
  h.graph_attr.update(rank='same',color='#cecdc9',style='filled',**label_cluster)
  g.subgraph(h)
  if i > 0:
    edges_cluster.append(
      (
        (nodes_by_term[i-1][0],nodes_by_term[i][0]),
        {'ltail': 'cluster_'+str(i-1),'lhead': name_now,'style': 'invis'}
      )
    )

g = add_edges(g, edges_cluster)
g = add_edges(g, edges)

font = 'Helvetica'
node_color = '#5e5d87'
styles = {
  'graph': {
    # 'label': 'Prerequisite structure',
    'fontsize': '16',
    'fontcolor': 'black',
    'fontname': font,
    'bgcolor': 'white',
    'rankdir': 'LR',
  },
  'nodes': {
    'fontname': font,
    'shape': 'rectangle',
    'fontcolor': node_color,
    'color': 'white',
    'style': 'filled',
    'fillcolor': 'white',
  },
  'edges': {
    'style': 'dashed',
    'color': 'black',
    'fontname': font,
    'fontsize': '12',
    'fontcolor': 'black',
    'penwidth': '3',
    'arrowsize': '1',
    'arrowhead': 'normal',
  }
}


g = apply_styles(g, styles)
g.render('img/'+sys.argv[1].split('.')[0])
g.view()

#### legend ####
# gl = digraph(engine='dot')
# gl.node('pr','pre-requisite')
# gl.node('cr','co-requisite')
# gl.node('pcr','pre- or co-requisite')
# gl.edge()

from graphviz import Source
temp = """
digraph {
  rankdir=LR
  node [shape=plaintext]
  fontname=Helvetica
  subgraph cluster_01 {
    label=Legend
    key [label=<<table border="0" cellpadding="2" cellspacing="0" cellborder="0">
      <tr><td align="right" port="i1">prerequisite  </td></tr>
      <tr><td align="right" port="i2">corequisite  </td></tr>
      <tr><td align="right" port="i3">con-prerequisite  </td></tr>
      </table>>, fontname=Helvetica]
    key2 [label=<<table border="0" cellpadding="2" cellspacing="0" cellborder="0">
      <tr><td port="i1">&nbsp;</td></tr>
      <tr><td port="i2">&nbsp;</td></tr>
      <tr><td port="i3">&nbsp;</td></tr>
      </table>>, fontname=Helvetica]
    key:i1:e -> key2:i1:w [penwidth=3]
    key:i2:e -> key2:i2:w [dir="both",penwidth=3]
    key:i3:e -> key2:i3:w [style=dashed,penwidth=3]
  }
}
"""
gl = Source(temp, filename="img/legend", format="svg")
gl.render('img/'+sys.argv[1].split('.')[0]+'_legend')
# gl.view()


import re
from smume.utils import term_sort_key, normalize_categories

def catalog_to_module_name(catalog):
    """
    Normalize a human-friendly catalog string to a valid Python module name.
    Examples:
        "ME 2024–25" -> "_2024_25"
        "2024–25"    -> "_2024_25"
        "2024-2025"  -> "_2024_25"
    """
    match = re.search(r"(\d{4})\D+(\d{2,4})", catalog)
    if match:
        start_year, end_year = match.groups()
        end_year = end_year[-2:]  # get last 2 digits
        return f"_{start_year}_{end_year}"
    raise ValueError(f"Unrecognized catalog format: {catalog}")


import importlib

class GenericPlan:
    """
    A plan representing the default recommended sequence for a given catalog year.
    Stores the initial term layout as defined in the curriculum module.
    """

    def __init__(self, catalog):
        if isinstance(catalog, str):
            module_name = catalog_to_module_name(catalog)
            mod = importlib.import_module(f"smume.curricula.{module_name}")
            self.curriculum = mod.curriculum
            self.catalog = catalog
        else:
            self.curriculum = catalog
            self.catalog = str(catalog)
        self.name = self.curriculum.name
        self.notes_generic = []

    @property
    def courses(self):
        """
        Returns all courses with a term assigned.
        This includes courses that are not yet completed.
        """
        return [course for course in self.courses_all if course.term]

    @property
    def courses_all(self):
        """
        Returns all courses in the curriculum.
        """
        return self.curriculum.courses.values()

    @property
    def course_terms(self):
        """
        Returns a mapping of course names to their assigned terms.
        """
        return {
            name: course.term for name, course in self.curriculum.courses.items()
        }

    @property
    def course_categories(self):
        """
        Returns a dictionary mapping category names to lists of courses.
        A course may appear in multiple categories.
        """
        from collections import defaultdict
        category_map = defaultdict(list)
        for course in self.courses:
            for category in course.categories:
                category_map[category].append(course)
        return dict(category_map)

    def get_term(self, course_name):
        return self.course_terms.get(course_name)

    def set_term(self, course_name, term):
        if course_name in self.course_terms:
            self.course_terms[course_name] = term
        if course_name in self.curriculum.courses:
            self.curriculum.courses[course_name].term = term
    
    def add_course(self, course_name, term):
        """
        Adds a course to the plan with a specific term.
        """
        if course_name not in self.curriculum.courses:
            raise ValueError(f"Course '{course_name}' not found in curriculum.")
        
        self.set_term(course_name, term)
    
    def remove_course(self, course_name):
        """
        Removes a course from the plan.
        """
        if course_name not in self.curriculum.courses:
            raise ValueError(f"Course '{course_name}' not found in curriculum.")
        
        self.set_term(course_name, None)
    
    def substitute_course(self, old_name, new_name):
        """
        Substitutes an old course with a new one.
        """
        if old_name not in self.curriculum.courses:
            raise ValueError(f"Course '{old_name}' not found in curriculum.")
        if new_name not in self.curriculum.courses:
            raise ValueError(f"Course '{new_name}' not found in curriculum.")

        old_course = self.curriculum.courses[old_name]
        new_course = self.curriculum.courses[new_name]

        # Copy term and other attributes from the old course
        new_course.term = old_course.term
        new_course.completed = old_course.completed
        new_course.critical_path = old_course.critical_path

        # Remove the old course from the plan
        old_course.term = None

    def switch_writing_intensive(self, old_name, new_name):
        """
        Substitutes one writing intensive course for another and adds the non-writing intensive version of the old course.
        """
        if old_name not in self.curriculum.courses:
            raise ValueError(f"Course '{old_name}' not found in curriculum.")
        if new_name not in self.curriculum.courses:
            raise ValueError(f"Course '{new_name}' not found in curriculum.")

        old_course = self.curriculum.courses[old_name]
        new_course = self.curriculum.courses[new_name]

        # Ensure both courses are writing intensive
        if not old_course.writing_intensive or not new_course.writing_intensive:
            raise ValueError("Both courses must be writing intensive.")

        # Find the non-writing intensive version of the old course
        non_writing_course_name = old_name.replace("W", "")
        if non_writing_course_name not in self.curriculum.courses:
            raise ValueError(f"Non-writing intensive version '{non_writing_course_name}' not found in curriculum.")
        non_writing_course = self.curriculum.courses[non_writing_course_name]

        # Add the non-writing intensive course to the plan by setting its term
        if old_course.term:
            self.set_term(non_writing_course_name, old_course.term)
        else:
            raise ValueError(f"Old course '{old_name}' does not have a term assigned. Cannot set term for non-writing intensive course.")
        
        # Remove the non-writing intensive version of the new course
        non_writing_course_name = new_name.replace("W", "")
        if non_writing_course_name in self.curriculum.courses:
            self.remove_course(non_writing_course_name)

        # Substitute the old course with the new one
        self.substitute_course(old_name, new_name)

    def apply_term_mapping(self, term_map: dict, check: bool = False):
        """
        Updates course terms based on a dictionary like:
        {
            "1F": ["ME 345", "ME 316"],
            "1S": ["ME 392"]
        }
        Only updates terms for the listed courses. Other courses remain unchanged.
        If check=True, returns a dictionary of dependency issues.
        """
        for term, course_list in term_map.items():
            for course_name in course_list:
                self.set_term(course_name, term)

        if check:
            return self.check_dependencies()
    
    def courses_by_term(self):
        grouped = {}
        for course in self.courses:
            grouped.setdefault(course.term, []).append(course)
        return grouped

    def check_dependencies(self):
        """
        Check for unmet prerequisites, corequisites, and coprerequisites
        in the course plan. Returns a dictionary of problems keyed by course name.
        """
        problems = {}

        term_order = sorted(
            {t for t in self.course_terms.values() if t is not None},
            key=term_sort_key
        )
    
        term_index = {term: i for i, term in enumerate(term_order)}

        for course in self.courses:
            course_term = self.course_terms.get(course.name)
            course_index = term_index.get(course_term, -1)

            unmet = {"prereq": [], "coreq": [], "coprereq": []}

            for pre in getattr(course, "prereqs", []):
                pre_term = self.course_terms.get(pre)
                if pre_term is None or term_index.get(pre_term, -1) >= course_index:
                    unmet["prereq"].append(pre)

            for co in getattr(course, "coreqs", []):
                co_term = self.course_terms.get(co)
                if co_term is None or term_index.get(co_term, -1) != course_index:
                    unmet["coreq"].append(co)

            for copre in getattr(course, "coprereqs", []):
                copre_term = self.course_terms.get(copre)
                if copre_term is None or term_index.get(copre_term, -1) > course_index:
                    unmet["coprereq"].append(copre)

            if any(unmet.values()):
                problems[course.name] = unmet

        return problems
    
    def print_dependency_issues(self):
        """
        Print any unmet prerequisites, corequisites, or coprerequisites.
        """
        issues = self.check_dependencies()
        if not issues:
            print("No dependency issues found.")
            return

        for course_name, unmet in issues.items():
            print(f"Course: {course_name}")
            if unmet["prereq"]:
                print(f"  - Unmet Prerequisites: {', '.join(unmet['prereq'])}")
            if unmet["coreq"]:
                print(f"  - Unmet Corequisites: {', '.join(unmet['coreq'])}")
            if unmet["coprereq"]:
                print(f"  - Unmet Coprerequisites: {', '.join(unmet['coprereq'])}")

    def check_category_requirements(self):
        """
        Check if the courses in the plan meet the category requirements defined in the curriculum.
        Returns a dictionary of unmet category requirements.
        """
        unmet_requirements = {}
        checking_methods = {
            "Writing Intensive": self.check_writing_intensive,
            "Number of Courses": self.check_number_of_courses,
            "Number of Credits": self.check_number_of_credits,
        }
        for category, requirements in self.curriculum.category_requirements.items():
            for requirement in requirements:
                # Call the appropriate checking method
                if "kind" not in requirement:
                    unmet_requirements[category] = "Missing 'kind' in requirement"
                    continue
                if requirement["kind"] not in checking_methods:
                    unmet_requirements[category] = f"Unknown requirement kind: {requirement['kind']}"
                    continue
                # Call the checking method for this requirement kind
                check_method = checking_methods[requirement["kind"]]
                result = check_method(requirement, category)
                if result:
                    if category not in unmet_requirements:
                        unmet_requirements[category] = {}
                    unmet_requirements[category][requirement.get("note", "No note")] = result
        return unmet_requirements
    
    def check_writing_intensive(self, requirement: dict, category: str, completed_only: bool = False):
        """
        Check if the plan meets the writing intensive requirements for a given category.
        """
        if completed_only:
            writing_courses = [c for c in self.courses if c.writing_intensive and c.completed and category in normalize_categories(c.categories, self.curriculum.valid_categories)]
        else:
            writing_courses = [c for c in self.courses if c.writing_intensive and category in normalize_categories(c.categories, self.curriculum.valid_categories)]
        if len(writing_courses) < requirement.get("number", 1):
            return f"At least {requirement['number']} writing intensive course(s) required in {category}."
        return None

    def check_number_of_courses(self, requirement: dict, category: str, completed_only: bool = False):
        """
        Check if the plan meets the number of courses required for a given category.
        """
        if completed_only:
            courses_in_category = [c for c in self.courses if c.completed and category in c.categories]
        else:
            courses_in_category = [c for c in self.courses if category in c.categories]
        if len(courses_in_category) < requirement.get("number", 1):
            return f"At least {requirement['number']} course(s) required in {category}."
        return None

    def check_number_of_credits(self, requirement: dict, category: str, completed_only: bool = False):
        """
        Check if the plan meets the number of credits required for a given category.
        """
        if completed_only:
            credits_in_category = sum(c.credits for c in self.courses if c.completed and category in c.categories)
        else:
            credits_in_category = sum(c.credits for c in self.courses if category in c.categories)
        if credits_in_category < requirement.get("number", 3):
            return f"At least {requirement['number']} credits required in {category}."
        return None
    
    def print_category_requirement_issues(self):
        """
        Print any unmet category requirements.
        """
        issues = self.check_category_requirements()
        if not issues:
            print("All category requirements met.")
            return

        for category, issue in issues.items():
            print(f"Category: {category}")
            for note, message in issue.items():
                print(f"  - Issue: {note}. {message}")
        print("Please review the curriculum for unmet category requirements.")
        return issues
    
    def category_requirement_status(self, category, completed_only=False):
        """
        Checks if the plan meets the category requirement for a given category.
        Returns True if the requirement is satisfied, a message if not.
        If completed_only is True, only considers completed courses.
        """
        if category not in self.curriculum.category_requirements:
            return None
        
        requirements = self.curriculum.category_requirements[category]

        for requirement in requirements:
            if "kind" not in requirement:
                return f"Missing requirement 'kind' for category '{category}'."
            if requirement["kind"] == "Writing Intensive":
                result = self.check_writing_intensive(requirement, category)
            elif requirement["kind"] == "Number of Courses":
                result = self.check_number_of_courses(requirement, category)
            elif requirement["kind"] == "Number of Credits":
                result = self.check_number_of_credits(requirement, category)
            else:
                return False
            
            if result:
                return result
        
        return True
    
    def fraction_of_category_satisfied(self, category, requirement_kind="Number of Credits", completed_only=False):
        """
        Returns the fraction of the category requirement that has been satisfied.
        For "Number of Courses", returns the fraction of courses completed.
        For "Number of Credits", returns the fraction of credits completed.
        For completed_only=True, only considers completed courses (fraction completed).
        For completed_only=False, considers all courses in the category (fraction planned).
        """
        if category not in self.curriculum.category_requirements:
            print(f"Category '{category}' not found in curriculum.")
            return None
        
        requirements = self.curriculum.category_requirements[category]
        total_required = 0
        total_completed = 0
        total_planned = 0

        for requirement in requirements:
            if requirement["kind"] == requirement_kind:
                if requirement_kind == "Number of Courses":
                    total_required += requirement.get("number", 1)
                    total_completed += len([c for c in self.courses if category in c.categories and c.completed])
                    total_planned += len([c for c in self.courses if category in c.categories])
                elif requirement_kind == "Number of Credits":
                    total_required += requirement.get("number")
                    total_planned += sum(c.credits for c in self.courses if category in c.categories)
                    total_completed += sum(c.credits for c in self.courses if category in c.categories and c.completed)

        if total_required == 0:
            return None
        
        if completed_only:
            return total_completed / total_required
        else:
            return total_planned / total_required
        
    def add_generic_note(self, note):
        """
        Adds a note to the generic plan.
        Notes are not tied to specific courses or terms.
        """
        self.notes_generic.append(note)


from smume.student_plan import StudentPlan
from smume.graph_builder import build_graph
import os
import re

class MEPlanDocument:
    def __init__(self, student_plan: StudentPlan, output_dir=".", filename="ME-Plan-Document"):
        self.plan = student_plan
        self.output_dir = output_dir
        self.filename = filename
        self.report = None
        self.graph = None

        os.makedirs(self.output_dir, exist_ok=True)

    def generate_report(self):
        """Generate HTML report from the student plan."""
        from smume.report import Report
        report = Report(self.plan)
        self.report = report

    def generate_graph(self):
        """Generate SVG graph from the student plan."""
        graph = build_graph(self.plan)
        raw_svg = graph.pipe(format="svg").decode('utf-8')

        # Remove hardcoded width/height from <svg ...>
        raw_svg = re.sub(r'\s(width|height)="[^"]+"', '', raw_svg)
        raw_svg = re.sub(r'<svg', '<svg style="width:100%;height:auto;"', raw_svg, count=1)

        self.graph = raw_svg

    def create_combined_document(self, styles_external=False):
        """Combine HTML report and embedded SVG graph into a single HTML string."""
        if not self.report:
            self.generate_report()
        if not self.graph:
            self.generate_graph()

        svg_content = f'<div class="graph-container">{self.graph}</div>'

        combined_html = f"""
        <html>
        <head><meta charset="utf-8"><title>Student Plan</title>
        {self.css_styles()}
        </head>
        <body>
            <div class="container">
                <h1 class="my-4">Student Plan for {self.plan.student_name}</h1>
                <div class="report-container">
                    {self.report.generate_html()}
                </div>
                <h2>3. Flowchart</h2>
                {svg_content}
            </div>
        </body>
        </html>
        """

        return combined_html
    
    def save_combined_document(self, styles_external=False):
        """Save the combined HTML document to a file."""
        combined_html = self.create_combined_document(styles_external=styles_external)
        combined_path = os.path.join(self.output_dir, f"{self.filename}.html")
        with open(combined_path, "w") as f:
            f.write(combined_html)
        return combined_path

    def css_styles(self):
        """Return CSS styles for the report."""
        high_level_styles = """
        <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap@3.3.7/dist/css/bootstrap.min.css" integrity="sha384-BVYiiSIFeK1dGmJRAkycuHAHRg32OmUcww7on3RYdg4Va+PmSTsz/K68vbdEjh4u" crossorigin="anonymous">
        <style>
            body { font-family: Palatino, serif; font-size: 10pt; }
            .graph-container { margin-top: 20px; }
            h1, h2, h3, h4 { color: #333; font-size: 10pt; font-weight: bold; margin-top: 20px; }
            h1 { font-size: 12pt; }
            ul { list-style-type: none; padding: 0; }
            li { margin: 2px 0; }
            ul ul li { margin-left: 7px; }

            @media print {
                body {
                    # margin: 0.5in;
                }
                .graph-container {
                    page-break-inside: avoid;
                    break-inside: avoid;
                    max-width: 100%;
                }
                svg {
                    page-break-inside: avoid;
                    break-inside: avoid;
                    width: 100%;
                    height: auto;
                }
                .report-container {
                    page-break-before: avoid;
                }
            }

            @page {
                size: Letter;
                margin: .25in .25in .25in .25in;
                @bottom-right {
                    content: "Page " counter(page) " of " counter(pages);
                    font-size: 10pt;
                }
            }
        </style>
        """
        return high_level_styles + self.report.css_styles()
    
    def export_combined_document_pdf(self):
        """Export the combined HTML document to PDF."""
        # Use WeasyPrint to convert HTML to PDF. Warning: This requires the WeasyPrint library to be installed. 
        print("Warning: Exporting to PDF requires WeasyPrint to be installed. See https://doc.courtbouillon.org/weasyprint/stable/first_steps.html#installation.")
        from weasyprint import HTML
        combined_html = self.create_combined_document(styles_external=False)
        output_pdf_path = os.path.join(self.output_dir, f"{self.filename}.pdf")
        print(f"Exporting combined document to PDF: {output_pdf_path}")
        HTML(string=combined_html).write_pdf(output_pdf_path)
        return output_pdf_path


def term_sort_key(term):
        """
        Custom sort key for terms in the format 'YYYY-F', 'YYYY-S', etc.
        Converts terms to a tuple of (year, season_order) for sorting.
        """
        import re
        match = re.match(r"(\d+)-([A-Za-z]+)", term)
        if not match:
            return (term,)
        year, season = match.groups()
        season_order = {'S': 0, 'Su': 1, 'Su1': 2, 'Su2': 3, 'F': 4}
        return (int(year), season_order.get(season, 99))

def normalize_categories(categories, valid_categories):
    if isinstance(categories, str):
        categories = [categories]
    for cat in categories:
        if cat not in valid_categories:
            raise ValueError(f"Invalid course category: {cat}. Must be one of: {', '.join(valid_categories.keys())}")
    return [valid_categories[c] for c in categories if c in valid_categories]


# styles.py

FONT = 'Helvetica'
NODE_COLOR = '#5e5d87'

styles = {
    'graph': {
        'fontsize': '16',
        'fontname': FONT,
        'bgcolor': 'white',
        'rankdir': 'LR',
    },
    'nodes': {
        'fontname': FONT,
        'shape': 'rectangle',
        'fontcolor': NODE_COLOR,
        'color': 'white',
        'style': 'filled',
        'fillcolor': 'white',
    },
    'edges': {
        'style': 'dashed',
        'color': 'black',
        'fontname': FONT,
        'fontsize': '12',
        'fontcolor': 'black',
        'penwidth': '3',
        'arrowsize': '1',
        'arrowhead': 'normal',
    }
}


from smume.course_model import Course
from smume.utils import normalize_categories

class Curriculum:
    """
    Container for a curriculum's set of courses. Defines structure, not term plans.
    """

    def __init__(self, name: str, categories_def: dict = None, ):
        self.name = name
        self.courses = {}
        self.define_categories(categories_def or {})
        self.category_requirements = {}
        self.explicit_category_requirements = {}
        self.implicit_category_requirements = {}
        self.update_implicit_category_requirements()
        self.merge_category_requirements()
        self.DTA_exemptions = {}

    def course(self, name, credits, categories=None, **kwargs):
        """
        Creates and registers a course in the curriculum.
        Does not set term or completion status.
        """
        categories = normalize_categories(categories, valid_categories=self.valid_categories)
        course = Course(name, credits, categories=categories, **kwargs)
        self.courses[name] = course
        return course
    
    def define_categories(self, categories_def: dict):
        """
        Defines the categories used in this curriculum.
        Categories are a dictionary mapping category names to their full names.
        """
        self.categories = categories_def.keys() or []
        self.category_names = {cat: category["name"] for cat, category in categories_def.items()}
        self.category_order = {category["name"]: category["order"] for category in categories_def.values()}
        self.valid_categories = {cat: cat for cat in self.categories}
        valid_aliases = {alias: cat for cat, aliases in categories_def.items() for alias in aliases}
        self.valid_categories.update(valid_aliases)

    def category_requirement(self, category: str, kind: str, number: int = None, note: str = None):
        """
        Defines a category requirement for the curriculum.
        """
        # Update explicit requirements
        if category not in self.explicit_category_requirements:
            self.explicit_category_requirements[category] = []
        self.explicit_category_requirements[category].append({
            "kind": kind,
            "number": number,
            "note": note
        })
        self.update_implicit_category_requirements() # Update implicit requirements
        self.merge_category_requirements()  # Merge explicit and implicit requirements
        return self

    def update_implicit_category_requirements(self):
        """
        Updates the implicit category requirements based on courses.
        If a category has no explicit Number of Credits requirements, we assume that all credits in that category must be completed.
        """
        implicit_reqs = {}
        for category in self.categories:
            explicit_requirements = self.explicit_category_requirements.get(category, [])
            if not any(req["kind"] == "Number of Credits" for req in explicit_requirements):
                # If no explicit Number of Credits requirements, assume all courses in category are required
                # However, don't double count credits if course is in multiple categories
                number_of_credits = sum(course.credits for course in self.courses.values() if category in course.categories)
                implicit_reqs[category] = {
                    "kind": "Number of Credits",
                    "number": number_of_credits,
                    "note": f"All courses in {category} category must be completed"
                }
        self.implicit_category_requirements = implicit_reqs
        return self.implicit_category_requirements

    def merge_category_requirements(self):
        """
        Merges explicit and implicit category requirements into a single dictionary (category_requirements).
        """
        merged_requirements = self.explicit_category_requirements.copy()
        for category, implicit_req in self.implicit_category_requirements.items():
            if category not in merged_requirements:
                merged_requirements[category] = []
            merged_requirements[category].append(implicit_req)
        self.category_requirements = merged_requirements
        return self.category_requirements
    
    def set_DTA_exemptions(self, name: str, course_exemptions: list):
        """
        Defines courses that are exempt from DTA requirements.
        Exemptions are a list of course names.
        """
        self.DTA_exemptions[name] = course_exemptions
        return self


# graph_builder.py

from graphviz import Digraph
from smume.utils import term_sort_key

def build_graph(plan, include_transfer_term=False, output_path=None, format="png"):
    graph = Digraph(format=format, engine='dot')
    graph.attr(rankdir='LR', newrank='true', compound='true', fontname='Palatino', fontsize='12')

    # Group courses by term
    courses_by_term = {}
    for course in plan.courses:
        if course.term:
            courses_by_term.setdefault(course.term, []).append(course)

    # Define category colors
    category_colors = {
        'C': '#f6e8c3',   # Core
        'MS': '#d5e8d4',  # Math and Science
        'GE': '#dae8fc',  # General Engineering
        'ME': '#f8cecc',  # Mechanical Engineering
        'O': '#e1d5e7',   # Other
    }

    # Define edge color palette (distinct, non-yellowish for clarity)
    edge_colors = [
        "#1f77b4", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd",
        "#8c564b", "#e377c2", "#7f7f7f", "#808000", "#17becf",
        "#393b79", "#637939", "#614100", "#843c39", "#7b4173",
        "#5254a3", "#6b6ecf", "#9c9ede", "#3182bd", "#31a354"
    ]

    sorted_terms = sorted(courses_by_term.keys(), key=term_sort_key)
    term_cluster_ids = []

    # Check for dependency violations if method exists
    violations = plan.check_dependencies() if hasattr(plan, 'check_dependencies') else {}

    # Create a subgraph for each term
    for idx, term in enumerate(sorted_terms):
        # Skip transfer term if not included explicitly
        if "transfer" in term.lower() and not include_transfer_term:
            continue
        cluster_name = f"cluster_{idx}"
        sub = Digraph(name=cluster_name)
        if type(plan).__name__ == 'StudentPlan':
            term_label = f""
        else:
            term_label = f"Term: "
        sub.attr(rank='same', style='filled', color='#cecdc9',
                 label=f"<<B>{term_label}{term}</B><BR/><FONT POINT-SIZE=\"10\" COLOR=\"#6e6d6a\">{sum(c.credits for c in courses_by_term[term])} cr</FONT>>")

        for course in courses_by_term[term]:
            if course.completed:
                fill_color = '#cecdc9'
            else:
                category_for_color = next((c for c in course.categories if c in category_colors), None)
                fill_color = category_colors.get(category_for_color, '#ffffff')
            label_text = f"{course.name}<font color=\"#6e6d6a\"><sub>  {course.credits} cr</sub></font>"
            if course.completed:
                label_text = f"✓ {label_text}"
            style_attrs = {
                'label': f"<{label_text}>",
                'style': 'filled',
                'fillcolor': fill_color,
                'shape': 'box',
                'fontsize': '10',
                'tooltip': getattr(course, 'note', ''),
            }
            is_violation_node = course.name in violations or any(
                course.name in deps
                for v in violations.values()
                for deps in [v.get('prereq', []) + v.get('coreq', []) + v.get('coprereq', [])]
            )
            if is_violation_node:
                style_attrs['color'] = 'red'
                style_attrs['penwidth'] = '3'
            style_attrs.update(course.styles)
            sub.node(course.name, **style_attrs)

        graph.subgraph(sub)
        term_cluster_ids.append((term, cluster_name))

    # Add invisible edges to control horizontal layout between clusters
    for i in range(1, len(term_cluster_ids)):
        prev_term, prev_cluster = term_cluster_ids[i - 1]
        curr_term, curr_cluster = term_cluster_ids[i]
        prev_courses = courses_by_term[prev_term]
        curr_courses = courses_by_term[curr_term]
        if prev_courses and curr_courses:
            graph.edge(prev_courses[0].name, curr_courses[0].name,
                       style='invis', ltail=prev_cluster, lhead=curr_cluster)

    # Add dependency edges with unique colors
    color_index = 0
    for course in plan.courses:
        for prereq in course.prereqs:
            edge_color = edge_colors[color_index % len(edge_colors)]
            color_index += 1
            completed = next((c for c in plan.courses if c.name == prereq and c.completed), None)
            is_violation = (
                course.name in violations
                and prereq in violations[course.name].get('prereq', [])
            )
            if is_violation:
                for node_name in [course.name, prereq]:
                    graph.node(node_name, color='red', penwidth='3')
            graph.edge(
                prereq, course.name,
                color='red' if is_violation else ('#aaaaaa' if completed else edge_color),
                style='solid',
                minlen='2',
                labelfloat='true'
            )
        for coreq in course.coreqs:
            edge_color = edge_colors[color_index % len(edge_colors)]
            color_index += 1
            completed = next((c for c in plan.courses if c.name == coreq and c.completed), None)
            is_violation = (
                course.name in violations
                and coreq in violations[course.name].get('coreq', [])
            )
            if is_violation:
                for node_name in [course.name, coreq]:
                    graph.node(node_name, color='red', penwidth='3')
            graph.edge(
                coreq, course.name,
                color='red' if is_violation else ('#aaaaaa' if completed else edge_color),
                style='bold',
                dir='both',
                minlen='2',
                labelfloat='true'
            )
        for coprereq in course.coprereqs:
            edge_color = edge_colors[color_index % len(edge_colors)]
            color_index += 1
            completed = next((c for c in plan.courses if c.name == coprereq and c.completed), None)
            is_violation = (
                course.name in violations
                and coprereq in violations[course.name].get('coprereq', [])
            )
            if is_violation:
                for node_name in [course.name, coprereq]:
                    graph.node(node_name, color='red', penwidth='3')
            graph.edge(
                coprereq, course.name,
                color='red' if is_violation else ('#aaaaaa' if completed else edge_color),
                style='dashed',
                minlen='2',
                labelfloat='true'
            )

    # Add legend as a subgraph using positioned nodes and edges
    legend = Digraph(name="cluster_legend")
    # small rank sep to keep legend compact
    legend.attr(label="Legend", fontsize="12", fontname="Palatino", style="dashed", color="#cecdc9", margin="0.2,0.2,0.2,0.2", ranksep="0.04", nodesep="0.0")

    # Create horizontally arranged edge-label pairs with reduced spacing

    legend.node('pr_edge', '', shape='point', width='0.2', height='0.2', margin='0', style='invis')
    legend.node('pr_label', 'prerequisite', shape='plaintext')
    legend.edge('pr_edge', 'pr_label', style='solid', color='black', minlen='2')

    legend.node('copr_edge', '', shape='point', width='0.2', height='0.2', margin='0', style='invis')
    legend.node('copr_label', 'con-prerequisite', shape='plaintext')
    legend.edge('copr_edge', 'copr_label', style='dashed', color='black', minlen='2')

    legend.node('cr_edge', '', shape='point', width='0.2', height='0.2', margin='0', style='invis')
    legend.node('cr_label', 'corequisite', shape='plaintext')
    legend.edge('cr_edge', 'cr_label', style='bold', color='black', dir='both', minlen='2') 

    # Force alignment: all edges on same rank, all labels on same rank
    same_rank_edges = Digraph()
    same_rank_edges.attr(rank='same', margin='.2', ranksep="0.08", nodesep="0.08")
    same_rank_edges.node('pr_edge')
    same_rank_edges.node('cr_edge')
    same_rank_edges.node('copr_edge')
    legend.subgraph(same_rank_edges)

    same_rank_labels = Digraph()
    same_rank_labels.attr(rank='same', margin='.2', ranksep="0.08", nodesep="0.08")
    same_rank_labels.node('pr_label')
    same_rank_labels.node('cr_label')
    same_rank_labels.node('copr_label')
    legend.subgraph(same_rank_labels)

    # Add an invisible edge to pull the legend toward bottom right
    if plan.courses:
        last_course = list(plan.courses)[-1]
        graph.edge(last_course.name, 'pr_edge', style='invis')
    graph.subgraph(legend)

    if output_path:
        graph.render(filename=output_path, format=format, cleanup=True)
    return graph


# colors.py

kelly_colors_hex = [
    "#FFB300",  # Vivid Yellow
    "#803E75",  # Strong Purple
    "#FF6800",  # Vivid Orange
    "#A6BDD7",  # Very Light Blue
    "#C10020",  # Vivid Red
    "#CEA262",  # Grayish Yellow
    "#817066",  # Medium Gray

    # The following don't work well for people with defective color vision
    "#007D34",  # Vivid Green
    "#F6768E",  # Strong Purplish Pink
    "#00538A",  # Strong Blue
    "#FF7A5C",  # Strong Yellowish Pink
    "#53377A",  # Strong Violet
    "#FF8E00",  # Vivid Orange Yellow
    "#B32851",  # Strong Purplish Red
    "#F4C800",  # Vivid Greenish Yellow
    "#7F180D",  # Strong Reddish Brown
    "#93AA00",  # Vivid Yellowish Green
    "#593315",  # Deep Yellowish Brown
    "#F13A13",  # Vivid Reddish Orange
    "#232C16",  # Dark Olive Green
]


digraph {
	graph [bgcolor=white fontcolor=black fontname=Helvetica fontsize=16 rankdir=LR]
	node [color=white fillcolor=white fontcolor="#5e5d87" fontname=Helvetica shape=rectangle style=filled]
	edge [arrowhead=normal arrowsize=1 color=black fontcolor=black fontname=Helvetica fontsize=12 penwidth=3 style=dashed]
	compound=true newrank=true
	subgraph cluster_0 {
		graph [color="#cecdc9" label=<<B>term: 1F</B><BR/><FONT POINT-SIZE="10" COLOR="#6e6d6a">9 cr</FONT>> rank=same style=filled]
		"PHY 171+L" [label=<<font color="#6e6d6a"><sub>5</sub></font>   PHY 171+L>]
		"MTH 171" [label=<<font color="#6e6d6a"><sub>4</sub></font>   MTH 171>]
	}
	subgraph cluster_1 {
		graph [color="#cecdc9" label=<<B>term: 1S</B><BR/><FONT POINT-SIZE="10" COLOR="#6e6d6a">12 cr</FONT>> rank=same style=filled]
		"PHY 172+L" [label=<<font color="#6e6d6a"><sub>5</sub></font>   PHY 172+L>]
		"GE 104" [label=<<font color="#6e6d6a"><sub>3</sub></font>   GE 104>]
		"MTH 172" [label=<<font color="#6e6d6a"><sub>4</sub></font>   MTH 172>]
	}
	subgraph cluster_2 {
		graph [color="#cecdc9" label=<<B>term: 2F</B><BR/><FONT POINT-SIZE="10" COLOR="#6e6d6a">14 cr</FONT>> rank=same style=filled]
		"GE 204" [label=<<font color="#6e6d6a"><sub>3</sub></font>   GE 204>]
		"CHM 145+L" [label=<<font color="#6e6d6a"><sub>5</sub></font>   CHM 145+L>]
		"MTH 322" [label=<<font color="#6e6d6a"><sub>3</sub></font>   MTH 322>]
		"MTH 271" [label=<<font color="#6e6d6a"><sub>3</sub></font>   MTH 271>]
	}
	subgraph cluster_3 {
		graph [color="#cecdc9" label=<<B>term: 2S</B><BR/><FONT POINT-SIZE="10" COLOR="#6e6d6a">10 cr</FONT>> rank=same style=filled]
		"GE 205" [label=<<font color="#6e6d6a"><sub>3</sub></font>   GE 205>]
		"MTH 353" [label=<<font color="#6e6d6a"><sub>3</sub></font>   MTH 353>]
		"GE 206+7" [label=<<font color="#6e6d6a"><sub>4</sub></font>   GE 206+7>]
	}
	subgraph cluster_4 {
		graph [color="#cecdc9" label=<<B>term: 3F</B><BR/><FONT POINT-SIZE="10" COLOR="#6e6d6a">17 cr</FONT>> rank=same style=filled]
		"ME 315" [label=<<font color="#6e6d6a"><sub>3</sub></font>   ME 315>]
		"ME 316" [label=<<font color="#6e6d6a"><sub>1</sub></font>   ME 316>]
		"ME 303" [label=<<font color="#6e6d6a"><sub>3</sub></font>   ME 303>]
		"ME 300+L" [label=<<font color="#6e6d6a"><sub>4</sub></font>   ME 300+L>]
		"ME 340" [label=<<font color="#6e6d6a"><sub>3</sub></font>   ME 340>]
		"ME 345" [label=<<font color="#6e6d6a"><sub>3</sub></font>   ME 345>]
	}
	subgraph cluster_5 {
		graph [color="#cecdc9" label=<<B>term: 3S</B><BR/><FONT POINT-SIZE="10" COLOR="#6e6d6a">13 cr</FONT>> rank=same style=filled]
		"ME 308+9" [label=<<font color="#6e6d6a"><sub>4</sub></font>   ME 308+9>]
		"ME 370" [label=<<font color="#6e6d6a"><sub>3</sub></font>   ME 370>]
		"ME 302" [label=<<font color="#6e6d6a"><sub>3</sub></font>   ME 302>]
		"ME 350" [label=<<font color="#6e6d6a"><sub>3</sub></font>   ME 350>]
	}
	subgraph cluster_6 {
		graph [color="#cecdc9" label=<<B>term: 4F</B><BR/><FONT POINT-SIZE="10" COLOR="#6e6d6a">12 cr</FONT>> rank=same style=filled]
		"ME 430+L" [label=<<font color="#6e6d6a"><sub>3</sub></font>   ME 430+L>]
		"ME Elective I" [label=<<font color="#6e6d6a"><sub>3</sub></font>   ME Elective I>]
		"ME 498" [label=<<font color="#6e6d6a"><sub>3</sub></font>   ME 498>]
		"ME 410" [label=<<font color="#6e6d6a"><sub>3</sub></font>   ME 410>]
	}
	subgraph cluster_7 {
		graph [color="#cecdc9" label=<<B>term: 4S</B><BR/><FONT POINT-SIZE="10" COLOR="#6e6d6a">12 cr</FONT>> rank=same style=filled]
		"GE 359" [label=<<font color="#6e6d6a"><sub>3</sub></font>   GE 359>]
		"ME Elective III" [label=<<font color="#6e6d6a"><sub>3</sub></font>   ME Elective III>]
		"ME 499+L" [label=<<font color="#6e6d6a"><sub>3</sub></font>   ME 499+L>]
		"ME Elective II" [label=<<font color="#6e6d6a"><sub>3</sub></font>   ME Elective II>]
	}
	"PHY 171+L" -> "PHY 172+L" [lhead=cluster_1 ltail=cluster_0 style=invis]
	"PHY 172+L" -> "GE 204" [lhead=cluster_2 ltail=cluster_1 style=invis]
	"GE 204" -> "GE 205" [lhead=cluster_3 ltail=cluster_2 style=invis]
	"GE 205" -> "ME 315" [lhead=cluster_4 ltail=cluster_3 style=invis]
	"ME 315" -> "ME 308+9" [lhead=cluster_5 ltail=cluster_4 style=invis]
	"ME 308+9" -> "ME 430+L" [lhead=cluster_6 ltail=cluster_5 style=invis]
	"ME 430+L" -> "GE 359" [lhead=cluster_7 ltail=cluster_6 style=invis]
	"PHY 172+L" -> "ME 315" [color="#803e75" style=solid]
	"ME 315" -> "ME 316" [color="#ff6800" style=dashed]
	"ME 345" -> "ME 316" [color="#a6bdd7" style=dashed]
	"GE 205" -> "ME 308+9" [color="#c10020" style=solid]
	"GE 206+7" -> "ME 308+9" [color="#cea262" style=solid]
	"ME 345" -> "ME 370" [color="#817066" style=solid]
	"GE 205" -> "ME 370" [color="#007d34" style=solid]
	"MTH 353" -> "ME 370" [color="#f6768e" style=dashed]
	"GE 206+7" -> "ME 303" [color="#00538a" style=solid]
	"CHM 145+L" -> "ME 303" [color="#ff7a5c" style=solid]
	"GE 206+7" -> "ME 302" [color="#53377a" style=solid]
	"MTH 171" -> "GE 204" [color="#ff8e00" style=solid]
	"PHY 171+L" -> "GE 204" [color="#b32851" style=solid]
	"GE 204" -> "GE 205" [color="#f4c800" style=solid]
	"ME 340" -> "ME 430+L" [color="#7f180d" style=solid]
	"ME 308+9" -> "ME 430+L" [color="#93aa00" style=solid]
	"MTH 322" -> "ME 430+L" [color="#593315" style=solid]
	"GE 206+7" -> "ME 300+L" [color="#f13a13" style=solid]
	"ME 302" -> "ME 498" [color="#232c16" style=solid]
	"ME 370" -> "ME 498" [color="#ffb300" style=solid]
	"ME 430+L" -> "ME 498" [color="#803e75" style=dashed]
	"MTH 172" -> "PHY 172+L" [color="#ff6800" style=dashed]
	"MTH 171" -> "PHY 171+L" [color="#a6bdd7" style=dashed]
	"MTH 171" -> "GE 104" [color="#c10020" style=solid]
	"PHY 171+L" -> "GE 104" [color="#cea262" style=solid]
	"ME 498" -> "ME 499+L" [color="#817066" style=solid]
	"MTH 172" -> "MTH 322" [color="#007d34" style=solid]
	"CHM 145+L" -> "ME 340" [color="#f6768e" style=solid]
	"PHY 172+L" -> "ME 340" [color="#00538a" style=solid]
	"MTH 322" -> "ME 340" [color="#ff7a5c" style=dashed]
	"MTH 172" -> "MTH 271" [color="#53377a" style=solid]
	"MTH 172" -> "MTH 353" [color="#ff8e00" style=solid]
	"MTH 322" -> "ME 345" [color="#b32851" style=solid]
	"PHY 172+L" -> "ME 345" [color="#f4c800" style=solid]
	"MTH 171" -> "MTH 172" [color="#7f180d" style=solid]
	"GE 104" -> "ME 350" [color="#93aa00" style=solid]
	"GE 204" -> "GE 206+7" [color="#593315" style=solid]
	"GE 205" -> "ME 410" [color="#f13a13" style=solid]
	"MTH 322" -> "ME 410" [color="#232c16" style=solid]
}



#!/usr/bin/env python3

import sys
from graphviz import Digraph, Source
from graph_builder import build_graph
from smume.course_model import courses
from smume.styles import styles

def apply_styles(graph, styles):
    graph.graph_attr.update(styles.get('graph', {}))
    graph.node_attr.update(styles.get('nodes', {}))
    graph.edge_attr.update(styles.get('edges', {}))
    return graph

def add_embedded_legend(graph):
    with graph.subgraph(name="cluster_legend") as legend:
        legend.attr(label="Legend", style="dashed")
        legend.attr(rank="sink")  # Push to bottom
        legend.node("prereq_label", "prerequisite", shape="plaintext")
        legend.node("coreq_label", "corequisite", shape="plaintext")
        legend.node("coprereq_label", "con-prerequisite", shape="plaintext")
        legend.node("prereq_box", "", shape="box", style="filled", width="0.3", height="0.2")
        legend.node("coreq_box", "", shape="box", style="filled,dashed", width="0.3", height="0.2")
        legend.node("coprereq_box", "", shape="box", style="filled,dotted", width="0.3", height="0.2")
        legend.edge("prereq_label", "prereq_box", penwidth="3")
        legend.edge("coreq_label", "coreq_box", dir="both", penwidth="3")
        legend.edge("coprereq_label", "coprereq_box", style="dashed", penwidth="3")

def main():
    graph = build_graph()
    graph = apply_styles(graph, styles)
    add_embedded_legend(graph)
    graph.render('img/flowchart', view=True)

if __name__ == '__main__':
    main()


class Report:
    def __init__(self, plan):
        self.plan = plan
        self.title = "Plan Report"
        self.styles = {}  # optional: for future styling options
        self.include_categories = None  # optional filter

    def generate_html(self):
        """Generate a structured HTML report of courses by category hierarchy."""
        def all_courses_completed(data):
            for key, value in data.items():
                if hasattr(value, 'name'):
                    if not getattr(value, 'completed', False):
                        return False
                else:
                    if not all_courses_completed(value):
                        return False
            return True

        def render_hierarchy(level, data):
            html = ['<ul>']
            category_order = self.plan.curriculum.category_order
            category_names = self.plan.curriculum.category_names

            def fmt_percent(value):
                return f"{value:.0%}" if value is not None else None

            # First, collect and render all courses
            course_items = []
            category_items = []
            for key, value in sorted(data.items(), key=lambda item: category_order.get(item[0], 999)):
                if hasattr(value, 'name'):
                    course_items.append((key, value))
                else:
                    category_items.append((key, value))

            for key, value in course_items:
                title = f"{value.name}: {value.full_name}" if getattr(value, "full_name", None) else value.name
                css_class = "course-item" + (" course-completed" if getattr(value, 'completed', False) else "")
                html.append(f'<li class="{css_class}">{title} ({value.credits})</li>')

            for key, value in category_items:
                cat_name = category_names.get(key, key)
                status = self.plan.category_requirement_status(key)
                completed = all_courses_completed(value)
                status_text = ""

                if status is not None:
                    if status and completed:
                        status_text = " (Complete)"
                    elif status and not completed:
                        planned_fraction = self.plan.fraction_of_category_satisfied(key, completed_only=False)
                        completed_fraction = self.plan.fraction_of_category_satisfied(key, completed_only=True)
                        if planned_fraction is None and completed_fraction is None:
                            status_text = ""
                        elif planned_fraction is None:
                            status_text = f" (Completed: {fmt_percent(completed_fraction)})"
                        elif completed_fraction is None:
                            status_text = f" (Planned: {fmt_percent(planned_fraction)})"
                        else:
                            status_text = f" (Planned: {fmt_percent(planned_fraction)}, Completed: {fmt_percent(completed_fraction)})"
                    elif not status:
                        unsatisfied_info = getattr(self.plan, 'category_requirement_unsatisfied_info', lambda k: "")(key)
                        status_text = f" (not satisfied{': ' + unsatisfied_info if unsatisfied_info else ''})"

                html.append(f'<li class="category-group">{cat_name}{status_text}{render_hierarchy(level + 1, value)}</li>')

            html.append('</ul>')
            return "\n".join(html)

        html = ['<div class="report">']
        html.append(f'<h2>{self.title}</h2>')
        html.append('<h3>1. Notes</h3>\n\n<ul>')
        for note in self.plan.notes:
            html.append(f'<li>{note["timestamp"]}: {note["text"]}</li>')
        html.append('</ul>')
        html.append('<h3>2. Courses by Category</h3>')
        rendered = render_hierarchy(0, self.hierarchy)
        if rendered.startswith('<ul>') and rendered.endswith('</ul>'):
            rendered = rendered[len('<ul>'):-len('</ul>')]
        html.append('<ul class="report-columns">')
        html.append(rendered)
        html.append('</ul>')
        html.append('</div>')
        return "\n".join(html)
    
    def generate_text(self):
        """Generate a hierarchical text report of courses by category."""
        lines = [self.title, "=" * len(self.title)]
        
        def all_courses_completed(data):
            for key, value in data.items():
                if hasattr(value, 'name'):
                    if not getattr(value, 'completed', False):
                        return False
                else:
                    if not all_courses_completed(value):
                        return False
            return True

        def render_hierarchy(level, data):
            indent = "  " * level
            category_order = self.plan.curriculum.category_order
            category_names = self.plan.curriculum.category_names
            # First, collect courses and categories
            course_items = []
            category_items = []
            for key, value in sorted(data.items(), key=lambda item: category_order.get(item[0], 999)):
                if hasattr(value, 'name'):
                    course_items.append((key, value))
                else:
                    category_items.append((key, value))

            for key, value in course_items:
                title = f"{value.name}: {value.full_name}" if getattr(value, "full_name", None) else value.name
                completed_mark = " ✓" if getattr(value, 'completed', False) else ""
                lines.append(f"{indent}- {completed_mark} {title} ({value.credits})")

            for key, value in category_items:
                cat_name = category_names.get(key, key)
                status = self.plan.category_requirement_status(key)
                completed = all_courses_completed(value)
                status_text = ""
                if status is not None:
                    if status and completed:
                        status_text = " (Complete)"
                    elif status and not completed:
                        planned_fraction = self.plan.fraction_of_category_satisfied(key, completed_only=False)
                        completed_fraction = self.plan.fraction_of_category_satisfied(key, completed_only=True)
                        if planned_fraction is None and completed_fraction is None:
                            status_text = ""
                        elif planned_fraction is None:
                            status_text = f" (Completed: {completed_fraction:.0%})"
                        elif completed_fraction is None:
                            status_text = f" (Planned: {planned_fraction:.0%})"
                        else:
                            status_text = f" (Planned: {planned_fraction:.0%}, Completed: {completed_fraction:.0%})"
                    elif not status:
                        unsatisfied_info = getattr(self.plan, 'category_requirement_unsatisfied_info', lambda k: "")(key)
                        status_text = f" (not satisfied{': ' + unsatisfied_info if unsatisfied_info else ''})"
                lines.append(f"{indent}{cat_name}{status_text}:")
                render_hierarchy(level + 1, value)

        render_hierarchy(0, self.hierarchy)
        return "\n".join(lines)
    
    @property
    def hierarchy(self):
        """
        Return a hierarchical structure of courses by category.
        Items without categories are at the top level.
        Items with just one category are grouped under that category at the second level.
        Items with two categories are grouped under both categories at the third level. Etc.
        If `include_categories` is set, only those categories are included.
        """
        hierarchy = {}
        for course in self.plan.courses:
            if self.include_categories and not any(cat in self.include_categories for cat in course.categories):
                continue
            current_level = hierarchy
            for cat in course.categories:
                if cat not in current_level:
                    current_level[cat] = {}
                current_level = current_level[cat]
            current_level[course.name] = course
        return hierarchy
    
    def save(self, filepath):
        """Save the report to the given filepath based on file extension."""
        content = self.generate_html() if filepath.endswith(".html") else self.generate_text()
        with open(filepath, "w", encoding="utf-8") as f:
            f.write(content)

    def css_styles(self):
        """Return CSS styles for the report, including print formatting."""
        return """
        <style>
            body { font-family: Palatino, serif; }
            .report { margin: 20px; }
            .category-group { font-weight: bold; }
            .course-item { margin-left: 20px; font-weight: normal; }
            .course-completed::before { content: "✓ "; }

            ul.report-columns {
                column-count: 2;
                column-gap: 40px;
            }
        </style>
        """


import re
from smume.generic_plan import GenericPlan

class StudentPlan(GenericPlan):
    """
    A student-specific plan that extends a catalog-based GenericPlan with
    specific academic years and term overrides.
    """

    def __init__(self, catalog: str, start_year: int, start_term: str = "Fall", student_name: str = None, student_id: str = None, DTA: str = None):
        super().__init__(catalog)
        self.start_year = start_year
        self.start_term = start_term  # "Fall" or "Spring"
        self.student_course_terms = {}
        self._assign_specific_terms()
        self.student_name = student_name or "Student"
        self.student_id = student_id  # Optional student ID, needed for parsing unofficial transcripts
        self._DTA = DTA
        self.notes = []  # Store student-specific notes
    
    @property
    def DTA(self):
        """
        Returns the DTA status of the student plan.
        If DTA is set, it returns the DTA type (e.g., "AA-DTA", "AS-DTA").
        Otherwise, returns None.
        """
        return self._DTA
    
    @DTA.setter
    def DTA(self, value):
        """
        Sets the DTA status of the student plan.
        Accepts "AA-DTA" or "AS-DTA" to indicate the type of DTA.
        """
        if value not in ["AA-DTA", "AS-DTA"]:
            raise ValueError("DTA must be 'AA-DTA' or 'AS-DTA'")
        self._DTA = value
        # Exempt the student via DTA
        self.exempt_DTA()

    def exempt_DTA(self):
        """
        Exempts the student from DTA requirements.
        """
        if self._DTA:
            print(f"Exempting student from {self._DTA} requirements.")
            # If DTA is set, mark all DTA courses as completed. Check also for W versions of exempted courses.
            w_versions = [course_name + "W" for course_name in self.curriculum.DTA_exemptions.get(self._DTA, [])]
            for course_name in self.curriculum.DTA_exemptions.get(self._DTA, []) + w_versions:
                if course_name in self.curriculum.courses:
                    # If there is a W writing intensive version, switch it to the non-W version ... I don't think this is necessary
                    # if course_name.endswith("W"):
                    #     term = self.student_course_terms.get(course_name, None) # Get the term of the W version
                    #     if term is None:
                    #         continue
                    #     term = self.extract_year_and_semester(term) if term else None
                    #     self.remove_course_term(course_name)  # Remove the W version
                    #     course_name = course_name[:-1] # Remove the 'W'
                    #     self.set_course_term(course_name, term.year, term.semester)  # Add the non-W version back to the plan
                    # If the course is not already completed, mark it as completed
                    # Mark the course as completed
                    course = self.curriculum.courses[course_name]
                    course.set_completed(True)
                    print(f"  Marking {course_name} as completed due to DTA exemption.")
        else:
            print("No DTA requirements found to exempt.")
    
    def extract_year_and_semester(self, term_label):
        """
        Extracts the year and semester from a term label like '2025-F' or '2024-S'.
        Returns a tuple (year, semester).
        """
        match = re.match(r"(\d{4})-(\w+)", term_label)
        if match:
            year = int(match.group(1))
            semester = match.group(2)
            return year, semester
        else:
            raise ValueError(f"Invalid term label format: {term_label}")

    def set_course_term(self, course_name: str, year, semester):
        """
        Assigns a specific normalized term label like '2025-F' to the course.
        """
        term_label = self._normalize_term_label(year, semester)
        self.student_course_terms[course_name] = term_label
        self.set_term(course_name, term_label)

    def remove_course_term(self, course_name: str):
        """
        Removes the term assignment for a course, effectively removing it from the plan.
        """
        self.student_course_terms.pop(course_name, None)
        if course_name in self.curriculum.courses:
            self.curriculum.courses[course_name].term = None
        else:
            print(f"Course {course_name} not found in curriculum, cannot remove term assignment.")

    def _normalize_term_label(self, year, semester):
        """
        Normalize year and semester into a term label like '2025-F'.
        Accepts 2- or 4-digit years and maps full season names to abbreviations.
        """
        if isinstance(year, int):
            if year == 0:
                year = "0000"  # Special case for transfer terms
            else:
                if year < 100:  # Handle 2-digit years
                    year += 2000
                year = str(year)
        elif isinstance(year, str):
            if year == "0000":
                year = "0000"
            else:
                if len(year) == 2 and year.isdigit():
                    year = "20" + year
                elif not year.isdigit() or len(year) not in [2, 4]:
                    raise ValueError(f"Invalid year format: {year}")
        else:
            raise TypeError("Year must be int or string")

        season_map = {
            "F": "F", "Fall": "F",
            "S": "S", "Spring": "S",
            "Su": "Su", "Summer": "Su",
            "Su1": "Su1", "Summer1": "Su1",
            "Su2": "Su2", "Summer2": "Su2",
            "Transfer": "Transfer"
        }

        sem = season_map.get(semester)
        if sem is None:
            raise ValueError(f"Invalid semester/season: {semester}. Must be one of: F, S, Su, Su1, Su2, Transfer.")

        return f"{year}-{sem}"

    def mark_completed(self, course_name: str):
        """
        Marks a course as completed.
        """
        course = self.curriculum.courses[course_name]
        if course:
            course.set_completed(True)
    def _assign_specific_terms(self):
        """
        Converts generic plan term labels like '1F', '2S', etc. into
        specific year-term labels based on student start year and term.
        Only considers S and F terms in a yearly cycle.
        """
        term_order = ["S", "F"]
        start_index = term_order.index(self._normalize_term_label(self.start_year, self.start_term).split("-")[1])
        generic_terms = sorted(t for t in self.courses_by_term().keys() if t is not None)

        year = self.start_year
        season_index = start_index
        for generic_term in generic_terms:
            actual_term = f"{year}-{term_order[season_index]}"
            for course in self.courses_by_term()[generic_term]:
                course_name = course.name
                self.set_term(course_name, actual_term)
                if course_name in self.curriculum.courses:
                    self.curriculum.courses[course_name].term = actual_term
            season_index += 1
            if season_index >= len(term_order):
                season_index = 0
                year += 1

    def add_note(self, note: str, date: str):
        """
        Adds a note to the student plan.
        Notes are specific to the student and not tied to courses or terms.
        :param note: The note text.
        :param date: The date of the note in 'YYYY-MM-DD' format.
        """
        note_structured = {
            "text": note,
            "timestamp": date
        }
        self.notes.append(note_structured)

    def parse_html_transcript(self, file_path: str):
        """
        Parses an HTML transcript file and extracts course information.
        :param file_path: Path to the HTML transcript file.
        """
        from bs4 import BeautifulSoup
        with open(file_path, 'r', encoding='utf-8') as file:
            soup = BeautifulSoup(file, 'html.parser')

        rows = soup.find_all("tr")

        current_term = None

        for row in rows:
            # Find the current term from the previous h2 tag
            h2_tag = row.find_previous("h2")
            if h2_tag is not None:
                current_term = h2_tag.get_text(strip=True)

            cells = row.find_all("td")
            if len(cells) >= 6:
                course_name = cells[0].get_text(strip=True)
                title = cells[1].get_text(strip=True)
                letter_grade = cells[3].get_text(strip=True)
                credits = cells[4].get_text(strip=True)
                quality_points = cells[5].get_text(strip=True)

                # Filter: If credits and quality points don't parse to floats, discard the row
                if not (credits.replace('.', '', 1).isdigit() and quality_points.replace('.', '', 1).isdigit()):
                    continue

                # Filter: If course name contains ["TERM", "OVERALL"], discard the row
                if any(term.lower() in course_name.lower() for term in ["TERM", "OVERALL"]):
                    continue

                # Normalize course name (some transcripts may have variations like "COR100" instead of "COR 100" or "ME100" instead of "ME 100"). We may need to split after the letters prefix.
                course_name = re.sub(r'(\D+)(\d+)', r'\1 \2', course_name)  # Add space between letters and numbers
                course_name = re.sub(r'\s+', ' ', course_name)  # Normalize multiple spaces
                course_name = course_name.strip().upper()
                if course_name in self.curriculum.courses:
                    course = self.curriculum.courses[course_name]
                else:
                    # Check if it's actually a DTA, not a course at all
                    if "AA-DTA" in course_name:
                        self.DTA = "AA-DTA"
                        print(f"  Detected DTA: {self.DTA}")
                        continue
                    if "AS-DTA" in course_name:
                        self.DTA = "AS-DTA"
                        print(f"  Detected DTA: {self.DTA}")
                        continue
                    # Add course with category Other
                    self.curriculum.course(course_name, credits=float(credits) if credits else 0.0, categories=["O"])
                    course = self.curriculum.courses[course_name]
                course.letter_grade = letter_grade
                course.grade = float(quality_points) / float(credits) if float(credits) > 0 else 0
                course.title = title
                course.credits = int(float(credits)) if credits else 0
                course.quality_points = float(quality_points) if quality_points else 0.0
                if course.letter_grade not in ["F", "", "W", "IP", "AU", "I", "NC"]:
                    print(f"  Marking course {course_name} as completed.")
                    course.set_completed(True)

                if current_term:
                    # Attempt to parse year and semester from current_term
                    # Expecting something like "2023 Fall" or "2024 Spring" or "0000 Transfer" 
                    match = re.match(r"(\d{4})\s*(Fall|Spring|Summer|Su1|Su2|Su|Transfer)", current_term, re.IGNORECASE)
                    print(f"  Current term match parts: {match.groups() if match else None}")
                    if match:
                        # Handle transfer terms. Change the start_year to 0000
                        if match.group(2).lower() == "transfer":
                            self.start_year = 0
                            self.start_term = "Transfer"
                        year_raw = match.group(1)
                        semester_raw = match.group(2)
                        print(f"  Setting term for {course_name} to semester {semester_raw} of year {year_raw}.")
                        try:
                            term_label = self._normalize_term_label(year_raw, semester_raw)
                            self.student_course_terms[course_name] = term_label
                            self.set_term(course_name, term_label)
                        except Exception as e:
                            print(f"  Error setting term for {course_name}: {e}")
                            pass


from smume.curriculum import Curriculum
from smume.generic_plan import GenericPlan

curriculum = Curriculum("ME 2024–25")

# Define Categories

categories_def = {
    "C": {"name": "Core", "order": 0, "aliases": ["Core"]},
    "F": {"name": "Foundation", "order": 1, "aliases": ["Foundation"]},
    "Con": {"name": "Conversatio", "order": 2, "aliases": ["Conversatio"]},
    "Ora": {"name": "Ora et Labora", "order": 3, "aliases": ["Ora et Labora"]},
    "MS": {"name": "Math and Science", "order": 4, "aliases": ["Math and Science"]},
    "GE": {"name": "General Engineering", "order": 5, "aliases": ["General Engineering"]},
    "ME": {"name": "Mechanical Engineering", "order": 6, "aliases": ["Mechanical Engineering"]},
    "O": {"name": "Other", "order": 7, "aliases": ["Other"]},
}
curriculum.define_categories(categories_def)

# Define the Courses

## Core Courses
curriculum.course('COR 100', 4, categories=["C", "F"], critical_path=False, full_name="First-Year Seminar")
curriculum.course('COR 110', 3, categories=["C", "F"], critical_path=False, full_name="Religious Studies")
curriculum.course('COR 120', 4, categories=["C", "F"], critical_path=True, full_name="Critical Reading and Writing")
curriculum.course('COR 210', 3, categories=["C", "Con"], critical_path=False, full_name="Humanities")
curriculum.course('COR 220', 3, categories=["C", "Con"], critical_path=False, full_name="Social Sciences")
curriculum.course('COR 240', 3, categories=["C", "Con"], critical_path=False, full_name="Artistic and Creative Expression")
curriculum.course('COR 250', 3, categories=["C", "Con"], critical_path=False, full_name="Historical and Political Studies")
curriculum.course('COR 310', 3, categories=["C", "Ora"], critical_path=False, full_name="Community: The Call to Serve the Common Good")
curriculum.course('COR 320', 3, categories=["C", "Ora"], critical_path=False, full_name="Hospitality and Openness to Others")
curriculum.course('COR 330', 3, categories=["C", "Ora"], critical_path=False, full_name="Stewardship: Responsible Use of Creation")
curriculum.course('COR 340', 3, categories=["C", "Ora"], critical_path=False, full_name="Ethics and the Dignity of Work")
curriculum.course('COR 210W', 4, categories=["C", "Con"], writing_intensive=True, critical_path=False, full_name="Humanities (Writing Intensive)")
curriculum.course('COR 220W', 4, categories=["C", "Con"], writing_intensive=True, critical_path=False, full_name="Social Sciences (Writing Intensive)")
curriculum.course('COR 240W', 4, categories=["C", "Con"], writing_intensive=True, critical_path=False, full_name="Artistic and Creative Expression (Writing Intensive)")
curriculum.course('COR 250W', 4, categories=["C", "Con"], writing_intensive=True, critical_path=False, full_name="Historical and Political Studies (Writing Intensive)")
curriculum.course('COR 310W', 4, categories=["C", "Ora"], writing_intensive=True, critical_path=False, full_name="Community: The Call to Serve the Common Good (Writing Intensive)")
curriculum.course('COR 320W', 4, categories=["C", "Ora"], writing_intensive=True, critical_path=False, full_name="Hospitality and Openness to Others (Writing Intensive)")
curriculum.course('COR 330W', 4, categories=["C", "Ora"], writing_intensive=True, critical_path=False, full_name="Stewardship: Responsible Use of Creation (Writing Intensive)")
curriculum.course('COR 340W', 4, categories=["C", "Ora"], writing_intensive=True, critical_path=False, full_name="Ethics and the Dignity of Work (Writing Intensive)")

## Math and Science
curriculum.course('MTH 171', 4, categories=["MS"], critical_path=True, full_name="Calculus I")
curriculum.course('PHY 171', 4, categories=["MS"], critical_path=True, full_name="Physics I").add_coprereq('MTH 171')
curriculum.course('PHY 171L', 1, categories=["MS"], critical_path=True, full_name="Physics I Laboratory").add_coreq('PHY 171')
curriculum.course('MTH 172', 4, categories=["MS"], critical_path=True, full_name="Calculus II").add_prereq('MTH 171')
curriculum.course('PHY 172', 4, categories=["MS"], critical_path=True, full_name="Physics II").add_coprereq('MTH 172').add_prereq('PHY 171')
curriculum.course('PHY 172L', 1, categories=["MS"], critical_path=True, full_name="Physics II Laboratory").add_coreq('PHY 172')
curriculum.course('CHM 145', 3, categories=["MS"], critical_path=True, full_name="Chemistry for Engineers", note="Chemistry for Engineers. Or CHM 141 (more credits).")
curriculum.course('MTH 353', 3, categories=["MS"], critical_path=True, full_name="Linear Algebra").add_prereq('MTH 172')

## General Engineering (GE)
curriculum.course('GE 104', 3, categories=["GE"], critical_path=True, full_name="Engineering Computing").add_prereq('MTH 171').add_prereq('PHY 171')
curriculum.course('GE 204', 3, categories=["GE"], critical_path=True, full_name="Statics").add_prereq('MTH 171').add_prereq('PHY 171')
curriculum.course('GE 205', 3, categories=["GE"], critical_path=True, full_name="Dynamics").add_prereq('GE 204')
curriculum.course('GE 206', 3, categories=["GE"], critical_path=True, full_name="Mechanics of Materials").add_prereq('GE 204')
curriculum.course('GE 207', 1, categories=["GE"], critical_path=True, full_name="Mechanics of Materials Laboratory").add_coprereq('GE 206')

## Mechanical Engineering (ME)
curriculum.course('ME 100', 1, categories=["ME"], critical_path=True, full_name="Mechanical Engineering Seminar")
curriculum.course('ME 201', 2, categories=["ME"], critical_path=True, full_name="Technical Communication").add_prereq('COR 120')
curriculum.course('ME 300', 3, categories=["ME"], critical_path=True, full_name="Manufacturing").add_prereq('GE 206')
curriculum.course('ME 300L', 1, categories=["ME"], critical_path=True, full_name="Manufacturing Laboratory").add_coprereq('ME 300')
curriculum.course('ME 303', 3, categories=["ME", "MS"], critical_path=True, full_name="Materials Science").add_prereq('GE 206').add_prereq('CHM 145')
curriculum.course('ME 316', 2, categories=["ME", "MS"], ms_credits=1, critical_path=True, full_name="Mechatronics and Instrumentation Laboratory").add_coprereq('ME 345')
curriculum.course('ME 340', 3, categories=["ME"], critical_path=True, full_name="Thermodynamics I").add_prereq('CHM 145').add_prereq('PHY 172').add_coprereq('MTH 172')
curriculum.course('ME 345', 3, categories=["ME"], critical_path=True, full_name="Mechatronics").add_prereq('MTH 172').add_prereq('PHY 172')
curriculum.course('ME 302', 3, categories=["ME"], critical_path=True, full_name="Machine Design").add_prereq('GE 206')
curriculum.course('ME 308', 3, categories=["ME"], critical_path=True, full_name="Fluid Mechanics I").add_prereq('GE 205').add_prereq('GE 206').add_prereq('MTH 172')
curriculum.course('ME 309', 2, categories=["ME", "MS"], ms_credits=1, critical_path=True, full_name="Fluid Mechanics Laboratory").add_coreq('ME 308')
curriculum.course('ME 350', 3, categories=["ME"], critical_path=True, full_name="Parametric Solid Modeling")
curriculum.course('ME 370', 3, categories=["ME"], critical_path=True, full_name="System Dynamics and Control").add_prereq('ME 345').add_prereq('GE 205').add_coprereq('MTH 353')
curriculum.course('ME 430', 3, categories=["ME"], critical_path=True, full_name="Heat Transfer").add_prereq('ME 340').add_prereq('ME 308').add_coreq('ME 430L')
curriculum.course('ME 430L', 2, categories=["ME", "MS"], ms_credits=1, critical_path=True, full_name="Heat Transfer Laboratory")
curriculum.course('ME 498', 3, categories=["ME"], critical_path=True, full_name="Senior Design I").add_prereq('ME 302').add_prereq('ME 370').add_coprereq('ME 430')
curriculum.course('ME 499', 3, categories=["ME"], critical_path=True, full_name="Senior Design II").add_prereq('ME 498')
curriculum.course('ME El. 1', 3, categories=["ME"], critical_path=True, full_name="Mechanical Engineering Elective 1")
curriculum.course('ME El. 2', 3, categories=["ME"], critical_path=True, full_name="Mechanical Engineering Elective 2")
curriculum.course('ME El. 3', 3, categories=["ME"], critical_path=True, full_name="Mechanical Engineering Elective 3")

# Define Category Requirements

# Core Requirements

## Total Credits (Have to be explicit because we defined all the W variants)
curriculum.category_requirement("C", kind="Number of Credits", number=31)
curriculum.category_requirement("F", kind="Number of Credits", number=11)
curriculum.category_requirement("Con", kind="Number of Credits", number=13)
curriculum.category_requirement("Ora", kind="Number of Credits", number=7)

## Specific Core Requirements
curriculum.category_requirement("Con", kind="Writing Intensive", number=1, note="One course must be writing intensive")
curriculum.category_requirement("Ora", kind="Writing Intensive", number=1, note="One course must be writing intensive")
curriculum.category_requirement("Ora", kind="Number of Courses", number=2, note="Two courses must be taken")

# Define DTA Exemptions
exemptions_DTA_AA = ["COR 100", "COR 120", "COR 210", "COR 220", "COR 240", "COR 250", "COR 310", "COR 320", "COR 330"]
curriculum.set_DTA_exemptions("AA-DTA", exemptions_DTA_AA)

# Assign Terms
generic_plan = GenericPlan(curriculum)
generic_plan.apply_term_mapping({
    "1F": ["MTH 171", "PHY 171", "PHY 171L", "ME 100", "COR 100"],
    "1S": ["MTH 172", "PHY 172", "PHY 172L", "GE 104", "COR 120"],
    "2F": ["GE 204", "CHM 145", "MTH 353", "ME 201", "COR 110"],
    "2S": ["GE 205", "GE 206", "GE 207", "ME 350", "COR 210"],
    "3F": ["ME 300", "ME 300L", "ME 303", "ME 316", "ME 345", "ME 340"],
    "3S": ["ME 302", "ME 308", "ME 309", "ME 370", "COR 220", "COR 240"],
    "4F": ["ME 430", "ME 430L", "ME 498", "ME El. 1", "COR 250W"],
    "4S": ["ME 499", "ME El. 2", "ME El. 3", "COR 310", "COR 340W"],
})
# generic_plan.print_dependency_issues() # Uncomment to check for dependency issues